/*
 * \file       simple_pkt_c_api.c
 * \brief      Reference CoreSight Trace Decoder : 
 * 
 * \copyright  Copyright (c) 2015, ARM Limited. All Rights Reserved.
 */

/* 
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 * 
 * 3. Neither the name of the copyright holder nor the names of its contributors 
 * may be used to endorse or promote products derived from this software without 
 * specific prior written permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 'AS IS' AND 
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */ 

/*
 * Example of using the library with the C-API
 *
 * Simple test program to print packets from a single trace source stream.
 * Hard coded configuration based on the Juno r1-1 test snapshot.
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <stdlib.h>

/* include the C-API library header */
#include "c_api/rctdl_c_api.h"

/* path to juno snapshot, relative to tests/bin/<plat>/<dbg|rel> build output dir */
#ifdef _WIN32
const char *default_path_to_snapshot = "..\\..\\..\\snapshots\\juno_r1_1\\";
#else
const char *default_path_to_snapshot = "../../../snapshots/juno_r1_1/";
#endif

/* trace data and memory file dump names */
const char *trace_data_filename = "cstrace.bin";
const char *memory_dump_filename = "kernel_dump.bin";
const rctdl_vaddr_t mem_dump_address=0xFFFFFFC000081000;

/* trace configuration structure - contains programmed register values of etmv4 hardware */
static rctdl_etmv4_cfg trace_config;

/* buffer to handle a packet string */
#define PACKET_STR_LEN 1024
static char packet_str[PACKET_STR_LEN];

/* decide if we decode & monitor, decode only or packet print */

typedef enum _test_op {
    TEST_PKT_PRINT,
    TEST_PKT_DECODE,
    TEST_PKT_DECODEONLY
} test_op_t;

static test_op_t op = TEST_PKT_PRINT; 

static uint8_t test_trc_id_override = 0x00;

static int print_data_array(const uint8_t *p_array, const int array_size, char *p_buffer, int buf_size)
{
    int chars_printed = 0;
    int bytes_processed;
    p_buffer[0] = 0;
    
    if(buf_size > 9)
    {
        /* set up the header */
        strcat(p_buffer,"[ ");
        chars_printed+=2;

        for(bytes_processed = 0; bytes_processed < array_size; bytes_processed++)
        {
           sprintf(p_buffer+chars_printed,"0x%02X ", p_array[bytes_processed]);
           chars_printed += 5;
           if((chars_printed + 5) > buf_size)
               break;
        }

        strcat(p_buffer,"];");
        chars_printed+=2;
    }
    else if(buf_size >= 4)
    {
        sprintf(p_buffer,"[];");
        chars_printed+=3;
    }
    return chars_printed;
}

/*  choose the operation to use for the test. */
static void process_cmd_line(int argc, char *argv[])
{
    int idx = 1;

    while(idx < argc)
    {
        if(strcmp(argv[idx],"-decode_only") == 0)
        {
            op = TEST_PKT_DECODEONLY;
        }
        else if(strcmp(argv[idx],"-decode") == 0)
        {
            op = TEST_PKT_DECODE;
        }
        else if(strcmp(argv[idx],"-id") == 0)
        {
            idx++;
            if(idx < argc)
            {
                test_trc_id_override = (uint8_t)(strtoul(argv[idx],0,0));
                printf("ID override = 0x%02X\n",test_trc_id_override);
            }
        }
        else 
            printf("Ignored unknown argument %s\n", argv[idx]);
        idx++;
    }
}

/* Callback function to process the packets in the stream - 
   simply print them out in this case 
 */
rctdl_datapath_resp_t etm_v4i_packet_handler(const rctdl_datapath_op_t op, const rctdl_trc_index_t index_sop, const rctdl_etmv4_i_pkt *p_packet_in)
{
    rctdl_datapath_resp_t resp = RCTDL_RESP_CONT;
    int offset = 0;

    switch(op)
    {
    default: break;
    case RCTDL_OP_DATA:
        sprintf(packet_str,"Idx:%ld; ", index_sop);
        offset = strlen(packet_str);
   
        /* got a packet - convert to string and use the libraries' message output to print to file and stdoout */
        if(rctdl_pkt_str(RCTDL_PROTOCOL_ETMV4I,(void *)p_packet_in,packet_str+offset,PACKET_STR_LEN-offset) == RCTDL_OK)
        {
            /* add in <CR> */
            if(strlen(packet_str) == PACKET_STR_LEN - 1) /* maximum length */
                packet_str[PACKET_STR_LEN-2] = '\n';
            else
                strcat(packet_str,"\n");

            /* print it using the library output logger. */
            rctdl_def_errlog_msgout(packet_str);
        }
        else
            resp = RCTDL_RESP_FATAL_INVALID_PARAM;  /* mark fatal error */
        break;

    case RCTDL_OP_EOT:
        sprintf(packet_str,"**** END OF TRACE ****\n");
        rctdl_def_errlog_msgout(packet_str);
        break;
    }

    return resp;
}


void etm_v4i_packet_monitor(  const rctdl_datapath_op_t op, 
                              const rctdl_trc_index_t index_sop, 
                              const rctdl_etmv4_i_pkt *p_packet_in,
                              const uint32_t size,
                              const uint8_t *p_data)
{
    int offset = 0;

    switch(op)
    {
    default: break;
    case RCTDL_OP_DATA:
        sprintf(packet_str,"Idx:%ld;", index_sop);
        offset = strlen(packet_str);
        offset+= print_data_array(p_data,size,packet_str+offset,PACKET_STR_LEN-offset);

        /* got a packet - convert to string and use the libraries' message output to print to file and stdoout */
        if(rctdl_pkt_str(RCTDL_PROTOCOL_ETMV4I,(void *)p_packet_in,packet_str+offset,PACKET_STR_LEN-offset) == RCTDL_OK)
        {
            /* add in <CR> */
            if(strlen(packet_str) == PACKET_STR_LEN - 1) /* maximum length */
                packet_str[PACKET_STR_LEN-2] = '\n';
            else
                strcat(packet_str,"\n");

            /* print it using the library output logger. */
            rctdl_def_errlog_msgout(packet_str);
        }
        break;

    case RCTDL_OP_EOT:
        sprintf(packet_str,"**** END OF TRACE ****\n");
        rctdl_def_errlog_msgout(packet_str);
        break;
    }
}

/* TBD: function to print generic packets. */
rctdl_datapath_resp_t gen_trace_elem_print(const rctdl_trc_index_t index_sop, const uint8_t trc_chan_id, const rctdl_generic_trace_elem *elem)
{
    rctdl_datapath_resp_t resp = RCTDL_RESP_CONT;
    int offset = 0;

    sprintf(packet_str,"Idx:%ld; TrcID:0x%02X; ", index_sop, trc_chan_id);
    offset = strlen(packet_str);

    if(rctdl_gen_elem_str(elem, packet_str+offset,PACKET_STR_LEN - offset) == RCTDL_OK)
    {
        /* add in <CR> */
        if(strlen(packet_str) == PACKET_STR_LEN - 1) /* maximum length */
            packet_str[PACKET_STR_LEN-2] = '\n';
        else
            strcat(packet_str,"\n");
    }
    else
    {
        strcat(packet_str,"Unable to create element string\n");
    }

    /* print it using the library output logger. */
    rctdl_def_errlog_msgout(packet_str);

    return resp;
}

/* hard coded values from snapshot .ini files */
void set_config_struct()
{
    trace_config.arch_ver   = ARCH_V8;
    trace_config.core_prof  = profile_CortexA;

    trace_config.reg_configr    = 0x000000C1;
    trace_config.reg_traceidr   = 0x00000010;   /* this is the trace ID -> 0x10, change this to analyse other streams in snapshot.*/

    if(test_trc_id_override != 0)
    {
        trace_config.reg_traceidr = (uint32_t)test_trc_id_override;
    }

    trace_config.reg_idr0   = 0x28000EA1;
    trace_config.reg_idr1   = 0x4100F403;
    trace_config.reg_idr2   = 0x00000488;
    trace_config.reg_idr8   = 0x0;
    trace_config.reg_idr9   = 0x0;
    trace_config.reg_idr10  = 0x0;
    trace_config.reg_idr11  = 0x0;
    trace_config.reg_idr12  = 0x0;
    trace_config.reg_idr13  = 0x0;
}

#define INPUT_BLOCK_SIZE 1024

/* process buffer until done or error */
rctdl_err_t process_data_block(dcd_tree_handle_t dcd_tree_h, int block_index, uint8_t *p_block, const int block_size)
{
    rctdl_err_t ret = RCTDL_OK;
    uint32_t bytes_done = 0;
    rctdl_datapath_resp_t dp_ret = RCTDL_RESP_CONT;
    uint32_t bytes_this_time = 0;

    while((bytes_done < (uint32_t)block_size) && (ret == RCTDL_OK))
    {
        if(RCTDL_DATA_RESP_IS_CONT(dp_ret))
        {
            dp_ret = rctdl_dt_process_data(dcd_tree_h, 
                                RCTDL_OP_DATA,
                                block_index+bytes_done,
                                block_size-bytes_done,
                                ((uint8_t *)p_block)+bytes_done,
                                &bytes_this_time);
            bytes_done += bytes_this_time;
        }
        else if(RCTDL_DATA_RESP_IS_WAIT(dp_ret))
        {
            dp_ret = rctdl_dt_process_data(dcd_tree_h, RCTDL_OP_FLUSH,0,0,NULL,NULL);
        }
        else
            ret = RCTDL_ERR_DATA_DECODE_FATAL; /* data path responded with an error - stop processing */
    }
    return ret;
}

int process_trace_data(FILE *pf)
{
    rctdl_err_t ret = RCTDL_OK;
    dcd_tree_handle_t dcdtree_handle = C_API_INVALID_TREE_HANDLE;
    int data_done = 0;
    uint8_t data_buffer[INPUT_BLOCK_SIZE];
    rctdl_trc_index_t index = 0;
    size_t data_read;
    char mem_file_path[512];

    /*  Create a decode tree for this source data.
        source data is frame formatted, memory aligned from an ETR (no frame syncs) so create tree accordingly 
    */
    dcdtree_handle = rctdl_create_dcd_tree(RCTDL_TRC_SRC_FRAME_FORMATTED, RCTDL_DFRMTR_FRAME_MEM_ALIGN);

    if(dcdtree_handle != C_API_INVALID_TREE_HANDLE)
    {
        /* populate the ETMv4 configuration structure */
        set_config_struct();

        if(op == TEST_PKT_PRINT) /* packet printing only */
        {
            /* Create a packet processor on the decode tree for the ETM v4 configuration we have. 
                We need to supply the configuration, and a packet handling callback.
            */
            ret = rctdl_dt_create_etmv4i_pkt_proc(dcdtree_handle,&trace_config,&etm_v4i_packet_handler);
        }
        else
        {
            /* Full decode - need decoder, and memory dump */

            /* create the packet decoder and packet processor pair */
            ret = rctdl_dt_create_etmv4i_decoder(dcdtree_handle,&trace_config);
            if(ret == RCTDL_OK)
            {
                /* attach the generic trace element output callback */
                ret = rctdl_dt_set_gen_elem_outfn(dcdtree_handle,gen_trace_elem_print);
                if((op != TEST_PKT_DECODEONLY) && (ret == RCTDL_OK))
                {
                     ret = rctdl_dt_attach_etmv4i_pkt_mon(dcdtree_handle, (uint8_t)(trace_config.reg_traceidr & 0xFF), etm_v4i_packet_monitor);
                }
            }

            /* trace data file path */
            strcpy(mem_file_path,default_path_to_snapshot);
            strcat(mem_file_path,memory_dump_filename);

            if(ret == RCTDL_OK)
            {
                /* create a memory file accessor */
                ret = rctdl_dt_add_binfile_mem_acc(dcdtree_handle,mem_dump_address,RCTDL_MEM_SPACE_ANY,mem_file_path);
            }
        }

        /* now push the trace data through the packet processor */
        while(!feof(pf) && (ret == RCTDL_OK))
        {
            /* read from file */
            data_read = fread(data_buffer,1,INPUT_BLOCK_SIZE,pf);
            if(data_read > 0)
            {
                /* process a block of data - any packets from the trace stream 
                   we have configured will appear at the callback 
                */
                ret = process_data_block(dcdtree_handle, 
                                index,
                                data_buffer,
                                data_read);
                index += data_read;
            }
            else if(ferror(pf))
                ret = RCTDL_ERR_FILE_ERROR;
        }

        /* no errors - let the data path know we are at end of trace */
        if(ret == RCTDL_OK)
            rctdl_dt_process_data(dcdtree_handle, RCTDL_OP_EOT, 0,0,NULL,NULL);

        /* dispose of the decode tree - which will dispose of any packet processors we created 
        */
        rctdl_destroy_dcd_tree(dcdtree_handle);
    }
    else
    {
        printf("Failed to create trace decode tree\n");
        ret = RCTDL_ERR_NOT_INIT;
    }
    return (int)ret;
}

int main(int argc, char *argv[])
{
    FILE *trace_data;
    char trace_file_path[512];
    int ret = 0;

    /* command line params */
    process_cmd_line(argc,argv);
    
    /* trace data file path */
    strcpy(trace_file_path,default_path_to_snapshot);
    strcat(trace_file_path,trace_data_filename);
    trace_data = fopen(trace_file_path,"rb");

    if(trace_data != NULL)
    {
        /* set up the logging in the library - enable the error logger, with an output printer*/
        ret = rctdl_def_errlog_init(RCTDL_ERR_SEV_INFO,1);
        
        /* set up the output - to file and stdout, set custom logfile name */
        if(ret == 0)
            ret = rctdl_def_errlog_config_output(C_API_MSGLOGOUT_FLG_FILE | C_API_MSGLOGOUT_FLG_STDOUT, "c_api_test.log");

        /* process the trace data */
        if(ret == 0)
            ret = process_trace_data(trace_data);

        /* close the data file */
        fclose(trace_data);
    }
    else
    {
        printf("Unable to open file %s to process trace data\n", trace_file_path);
        ret = -1;
    }
    return ret;
}
/* End of File simple_pkt_c_api.c */
